/*********************************************************************
 *
 * Copyright (c) 1998
 * 3Com/Palm Computing Division.  All rights reserved.
 *
 * Portions copyright (c) 1996-1997
 * Greg Hewgill.  All rights reserved.
 *
 * Portions copyright (c) 1995-1997
 * Illume Software, Inc.  All rights reserved.
 *
 * Portions copyright (c) 1995-1997
 * Bernd Schmidt.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *		This product includes software developed by 3Com and its
 *		contributors.
 *
 * 4. Neither 3Com nor the names of its contributors may be used to
 *    endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE 3COM AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 3COM OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 *******************************************************************/

#pragma once

#include "Switches.h"
#include "UAE.h"
#include "DragonBall.h"

#if HAS_PROFILING
#include "Profiling.h"			// gProfilingCounted
#endif


// Types.

// These two structs are used to control access to memory.  The
// first struct contains booleans which, when set to true, turn on
// address validation for the various ranges of memory.  The second
// struct contains booleans which, when set to true, prevent access
// by user applications to the various ranges of memory.

struct ValidateMemory
{
	bool				fDummyGet;
	bool				fDummySet;
	bool				fRegisterGet;
	bool				fRegisterSet;
	bool				fDRAMGet;
	bool				fDRAMSet;
	bool				fSRAMGet;
	bool				fSRAMSet;
	bool				fROMGet;
	bool				fROMSet;
};

struct ProtectMemory
{
	bool				fLowMemGet;
	bool				fLowMemSet;
	bool				fGlobalGet;
	bool				fGlobalSet;
	bool				fScreenGet;
	bool				fScreenSet;
	bool				fSRAMGet;
	bool				fSRAMSet;
	bool				fRegisterGet;
	bool				fRegisterSet;
};


// Constants.

const uaecptr			kDRAMBank_Start		= 0x00000000;
const uaecptr			kSRAMBank_Start		= hwrCardBase0;
const uaecptr			kROMBank_Start		= 0x10C00000;	// Accessed from Hardware_Reset
const uaecptr			k68681Bank_Start	= hwrDuartBase;
const uaecptr			kRegisterBank_Start	= hwr68328Base;



// Map some good names to some names compatible with the
// sources generated by GenCPU.

#define AddressBank				addrbank
#define Memory_GetLong			longget		// "get_long" instead?
#define Memory_GetWord			wordget		// ditto
#define Memory_GetByte			byteget		// ditto
#define Memory_SetLong			longput		// ditto
#define Memory_SetWord			wordput		// ditto
#define Memory_SetByte			byteput		// ditto
#define Memory_ValidAddress		valid_address
#define gMemory_Banks			mem_banks


// Macro to return the Dragonball address of the specified register

#define addressof(x)	(kRegisterBank_Start + offsetof(HwrDBallType, x))


// Macros for reading/writing Dragonball registers.

#define READ_REGISTER(reg)	\
	((sizeof(g68328Regs.reg) == 1) ? do_get_mem_byte(&g68328Regs.reg) :	\
	 (sizeof(g68328Regs.reg) == 2) ? do_get_mem_word(&g68328Regs.reg) :	\
									 do_get_mem_long(&g68328Regs.reg))

#define WRITE_REGISTER(reg, value)	\
	(sizeof(g68328Regs.reg) == 1) ? do_put_mem_byte(&g68328Regs.reg, value) :	\
	(sizeof(g68328Regs.reg) == 2) ? do_put_mem_word(&g68328Regs.reg, value) :	\
									do_put_mem_long(&g68328Regs.reg, value)

// Globals.

extern char*			gRAM_Memory;		// Accessed from CEmulatorApp::AskToSave

extern HwrDBallType		g68328Regs;			// Initialized from Hardware_Initialize
extern uae_u16			gLastTmr1Status;	// Initialized from Hardware_Initialize
extern uae_u16			gLastTmr2Status;	// Initialized from Hardware_Initialize
extern uae_u8			gPortDEdge;			// Initialized from Hardware_Initialize

extern unsigned long	gPreventAccessNotificationFlags;	// Headpatch::SysUIAppSwitch

extern ValidateMemory	gValidateMemory;
extern ProtectMemory	gProtectMemory;


// There are places within the emulator where we'd like to access low-memory
// and/or Dragonball registers.  If the PC happens to be in RAM, then
// the checks implied by the above booleans and switches will flag our
// access as an error.  Before making such accesses, create an instance
// of CEnableFullAccess to suspend and restore the checks.
//
// Since such accesses are typically "meta" accesses where the emulator is
// accessing memory outside the normal execution of an opcode, we also
// turn off the profiling variable that controls whether or not cycles
// spent accessing memory are counted.

class CEnableFullAccess
{
	public:
				CEnableFullAccess () :
					fValidateMemory (gValidateMemory),
					fProtectMemory (gProtectMemory)
#if HAS_PROFILING
					, fProfilingCounted (gProfilingCounted)
#endif
				{					
					memset (&gValidateMemory, 0, sizeof (gValidateMemory));
					memset (&gProtectMemory, 0, sizeof (gProtectMemory));
#if HAS_PROFILING
					gProfilingCounted = false;
#endif
				}

				~CEnableFullAccess ()
				{
					gValidateMemory = fValidateMemory;
					gProtectMemory = fProtectMemory;
#if HAS_PROFILING
					gProfilingCounted = fProfilingCounted;
#endif
				}

	private:
		ValidateMemory	fValidateMemory;
		ProtectMemory	fProtectMemory;
#if HAS_PROFILING
		bool			fProfilingCounted;
#endif
};


// Function prototypes.

class Memory
{
	public:
		static void		Initialize			(FILE* iRAMFile,
											 FILE* iROMFile,
											 uae_s32 iRAMSize);
		static void		Reset				(void);
		static void		Dispose				(bool iAlsoDisposeROM = true);
		static void		Save				(FILE* iMemoryFile);
		static void		InitializeBanks		(AddressBank& iBankInitializer,
											 uae_s32 iStartingBankIndex,
											 uae_s32 iNumberOfBanks);
		static void		InitializeROM		(FILE* iROMFile);
		static void		InitializeRAM		(FILE* iRAMFile, uae_s32 iRAMSize);
		static void		MapPhysicalMemory	(const void*, uae_u32);
		static void		UnmapPhysicalMemory	(const void*);
};
