/*********************************************************************
 *
 * Copyright (c) 1998
 * 3Com/Palm Computing Division.  All rights reserved.
 *
 * Portions copyright (c) 1996-1997
 * Greg Hewgill.  All rights reserved.
 *
 * Portions copyright (c) 1995-1997
 * Illume Software, Inc.  All rights reserved.
 *
 * Portions copyright (c) 1995-1997
 * Bernd Schmidt.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *		This product includes software developed by 3Com and its
 *		contributors.
 *
 * 4. Neither 3Com nor the names of its contributors may be used to
 *    endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE 3COM AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 3COM OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 *******************************************************************/

#include <windows.h>
#pragma hdrstop

#include "Platform.h"

#include "MMFMessaging.h"	// MMFSendData
#include "StringData.h"		// kErrorStrings
#include "TrapPatches.h"	// gCurrentApplicationName
#include "Settings.h"		// Settings.CommPort
#include "Dragonball.h"		// hwr328UBaudDivider, hwr328UBaudPrescaler
#include "CPU_REG.h"		// Software::BusError
#include "Gremlins.h"		// Gremlins_IsOn, Gremlins_GremlinNumber, Gremlins_EventNumber
#include "UAE_Utils.h"		// uae_strlen, uae_strcpy

#include "SysTraps.h"		// sysTrapSndDoCmd

#include "resource.h"


// Types.

struct SndCommandType		// !!! should use real type from Palm headers
{
	uae_u16	cmd;
	uae_u16	param1hi;
	uae_u16	param1lo;
	uae_u16	param2;
	uae_u16	param3;
};


// Queue used for keyboard and serial input/output.  This is the
// private, platform-dependent implementation used by TByteQueue.

// This class does not need a mutex because it assumes there is only
// one producer (who calls Put) and one consumer (who calls Get).

class TByteQueuePrivate
{
	public:
								TByteQueuePrivate	();
								~TByteQueuePrivate	();

		void					Put			(uae_u8 b);
		uae_u8					Get			(void);
		int						GetUsed		(void) { return fUsedCount; }
		int						GetFree		(void) { return fFreeCount; }
		bool					WaitForDataAvailable (int timeoutms);

	private:

		enum { QUEUE_SIZE = 8 };

		long					fUsedCount;
		long					fFreeCount;
		HANDLE					fDataAvailable;
		uae_u8					fBuffer[QUEUE_SIZE];
		int						fHead;
		int						fTail;
};


// Globals.

extern HINSTANCE	gInstance;			// In Emulator.cpp
extern HWND			gEmulatorWindow;	// In LCD.cpp
extern HANDLE		gCPUDelayEvent;		// In PilotCPU.cpp

bool				EnableSounds;
unsigned long		gPseudoTickcount;

HANDLE				gCommHandle;
TByteQueue			gCommReadQueue;
TByteQueue			gCommWriteQueue;

extern void CenterWindow (HWND);	// In LCD.cpp


// Private functions.

DWORD __stdcall CommRead (void*);
DWORD __stdcall CommWrite (void*);


// ---------------------------------------------------------------------------
//		 Platform::Initialize
// ---------------------------------------------------------------------------
// Initializes platform-dependent stuff.

void Platform::Initialize (void)
{
	// Open the serial port (creates gCommHandle)

	Platform::OpenSerialPort ();
}


// ---------------------------------------------------------------------------
//		 Platform::Reset
// ---------------------------------------------------------------------------

void Platform::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 Platform::ReportString
// ---------------------------------------------------------------------------
// Displays a stop alert that contains the specified message.

void Platform::ReportString (const unsigned char* iMessageString)
{
	// Shouldn't really be called on Windows -- no prevalent use of Pascal strings.
}


// ---------------------------------------------------------------------------
//		 Platform::ReportString
// ---------------------------------------------------------------------------
// Displays a stop alert that contains the specified message.

void Platform::ReportString (const char* iMessageString)
{
	(void) ::MessageBox (gEmulatorWindow, iMessageString,
						"Palm OS Emulator Error", MB_OK | MB_ICONERROR);
}



// ---------------------------------------------------------------------------
//		 Platform::ReportError
// ---------------------------------------------------------------------------
// Displays a stop alert that contains the specified message.

void Platform::ReportError (long iMessageID)
{
	// Get the error string.

	if (iMessageID >= kError_LastError)
		iMessageID = kError_DefaultError;


	// Display the dialog.

	Platform::ReportString (kErrorStrings[iMessageID]);
}


// ---------------------------------------------------------------------------
//		 Platform::ReportCriticalError
// ---------------------------------------------------------------------------
// Displays a stop alert that contains the specified error and the file name
// and line number where the error occurred.

void Platform::ReportCriticalError (long		iMessageID,
									long		iErrorID,
									const char*	iFileName,
									long		iLineNumber)
{
	// Get the error string.

	if (iMessageID >= kError_LastError)
		iMessageID = kError_DefaultError;

	// Try to determine the error if no error code was provided.

	DWORD	aErrorID = iErrorID;
	if (aErrorID == kError_NoError)
		aErrorID = GetLastError ();

	// Convert the error code to a string.

	char	aError[256];
	_itoa (aErrorID, aError, 10);

	// Convert the line number to a string.

	char	aLineNumber[256];
	_itoa (iLineNumber, aLineNumber, 10);

	// Set the error text.

	char	aFullMessage[256];

	sprintf (aFullMessage, "Error: %s\n\nError: %s\nFile: %s\nLine: %s",
		kErrorStrings[iMessageID], aError, iFileName, aLineNumber);

	// Display the dialog.

	Platform::ReportString (aFullMessage);
}


// ---------------------------------------------------------------------------
//		 Platform::Delay
// ---------------------------------------------------------------------------
// Delay for a time period appropriate for a sleeping 68328.

void Platform::Delay (void)
{
	// Delay 10 msecs.  Delaying by this amount pauses us 1/100 of a second,
	// which is the rate at which the device's tickcount counter increments.
	//
	// Wait on an event instead of just calling Sleep(10) so that another
	// thread can wake us up before our time.

	WaitForSingleObject (gCPUDelayEvent, 10);
}


// ---------------------------------------------------------------------------
//		 Platform::MessageBox
// ---------------------------------------------------------------------------
// Present a dialog containing a caller-supplied message and one or more of
// the following buttons: Continue, Debug, Reset.  Return which of the buttons
// was clicked on.

static const char*	gMessage;

static BOOL CALLBACK I_wish_I_had_lamda_functions (HWND hwnd, UINT msg,
												   WPARAM wparam, LPARAM lparam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			CenterWindow (hwnd);
			SetDlgItemText (hwnd, IDC_ERROR_MESSAGE, gMessage);
			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDOK:
				case IDC_RESET:
				case IDC_DEBUG:
					EndDialog (hwnd, LOWORD (wparam));
					break;
			}
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}

int Platform::ErrorDialog (const char* msg, int flags)
{
	gMessage = msg;

	UINT	dlgID = -1;
	switch (flags)
	{
		case kMsgContinue:
			dlgID = IDD_MSG_CONTINUE;
			break;

		case kMsgContinue + kMsgDebug:
			dlgID = IDD_MSG_CONTINUE_DEBUG;
			break;

		case kMsgContinue + kMsgDebug + kMsgReset:
			dlgID = IDD_MSG_CONTINUE_DEBUG_RESET;
			break;

		case kMsgContinue + kMsgReset:
			dlgID = IDD_MSG_CONTINUE_RESET;
			break;

		case kMsgDebug:
			dlgID = IDD_MSG_DEBUG;
			break;

		case kMsgDebug + kMsgReset:
			dlgID = IDD_MSG_DEBUG_RESET;
			break;

		case kMsgReset:
			dlgID = IDD_MSG_RESET;
			break;
	}

	int	button = DialogBox (gInstance, MAKEINTRESOURCE (dlgID),
							gEmulatorWindow, I_wish_I_had_lamda_functions);

	return	(button == IDOK) ? kMsgContinue :
			(button == IDC_DEBUG) ? kMsgDebug :
			kMsgReset;
}


// ---------------------------------------------------------------------------
//		 Platform::GetString
// ---------------------------------------------------------------------------

char* Platform::GetString (int id)
{
	int		amtRead;
	int		size = 256;		// Arbitrary starting size
	char*	result = NULL;

	// Allocate a buffer; load the string; loop if the buffer wasn't
	// big enough.

	while (1)
	{
		result = (char*) realloc (result, size);
		amtRead = ::LoadString(gInstance, id, result, size);
		if (amtRead < (size - 1))
			break;
		size += 256;	// Arbitrary amount by which to increment
	};

	return result;
}

// ---------------------------------------------------------------------------
//		 Platform::Cycle
// ---------------------------------------------------------------------------

void Platform::Cycle (void)
{
	// Nothing to do in Windows.
}


// ---------------------------------------------------------------------------
//		 Platform::GetMilliseconds
// ---------------------------------------------------------------------------

uae_u32 Platform::GetMilliseconds (void)
{
	return (uae_u32) GetTickCount();
}


// ---------------------------------------------------------------------------
//		 Platform::GetTime
// ---------------------------------------------------------------------------

void Platform::GetTime (long* hour, long* min, long* sec)
{
	SYSTEMTIME st;
	GetLocalTime(&st);

	*hour = st.wHour;
	*min = st.wMinute;
	*sec = st.wSecond;
}


// ---------------------------------------------------------------------------
//		 Platform::GetDate
// ---------------------------------------------------------------------------

void Platform::GetDate (long* year, long* month, long* day)
{
	SYSTEMTIME st;
	GetLocalTime(&st);

	*year = st.wYear;
	*month = st.wMonth;
	*day = st.wDay;
}


// ---------------------------------------------------------------------------
//		 Platform::HeadPatch
// ---------------------------------------------------------------------------

#if 0
int Platform::HeadPatch (uae_u16 iATrap)
{
	switch (iATrap)
	{
		case sysTrapSndDoCmd:
		{
			SndCommandType*	sc = (SndCommandType*) get_real_address(get_long(m68k_areg(regs, 7) + 4));
			if (do_get_mem_word(&sc->cmd) == 1)
			{
				if (EnableSounds)
				{
					Beep(do_get_mem_word(&sc->param1lo), do_get_mem_word(&sc->param2));
				}

				return true;
			}

			break;
		}
	}

	return false;
}
#endif


// ---------------------------------------------------------------------------
//		 Platform::ConnectToDebugger
// ---------------------------------------------------------------------------

uae_u32 Platform::ConnectToDebugger (void)
{
	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::SendPacket
// ---------------------------------------------------------------------------

uae_u32 Platform::SendPacket (const void* pkt, long size)
{
	return (uae_u32) MMFSendData (pkt, size);
}


// ---------------------------------------------------------------------------
//		 Platform::DisconnectFromDebugger
// ---------------------------------------------------------------------------

void Platform::DisconnectFromDebugger (void)
{
}


// ---------------------------------------------------------------------------
//		 Platform::ViewDrawLine
// ---------------------------------------------------------------------------

void Platform::ViewDrawLine (int xStart, int yStart, int xEnd, int yEnd)
{
}


// ---------------------------------------------------------------------------
//		 Platform::ViewDrawPixel
// ---------------------------------------------------------------------------

void Platform::ViewDrawPixel (int xPos, int yPos)
{
}


// ---------------------------------------------------------------------------
//		 Platform::OpenSerialPort
// ---------------------------------------------------------------------------

void Platform::OpenSerialPort (void)
{
	// Make sure the serial port is closed.

	CloseSerialPort ();

	// Create the comm handle used for serial communications.

	if (strnicmp (Settings.CommPort, "COM", 3) == 0)
	{
		gCommHandle = CreateFile (Settings.CommPort, GENERIC_READ | GENERIC_WRITE, 0,
								NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);

		if (gCommHandle == INVALID_HANDLE_VALUE)
		{
			char	buf[MAX_PATH];
			wsprintf (buf, "Communications port %s already in use", Settings.CommPort);
			::MessageBox (NULL, buf, "Palm OS Emulator", MB_OK);
			gCommHandle = NULL;
		}
	}

	if (gCommHandle)
	{
		// Initialize the serial port settings.

		DCB	dcb;
		dcb.DCBlength		= sizeof (dcb);
		GetCommState (gCommHandle, &dcb);

		dcb.BaudRate		= CBR_9600;
		dcb.fBinary			= TRUE;
		dcb.fParity			= FALSE;
		dcb.fOutxCtsFlow	= TRUE;
		dcb.fOutxDsrFlow	= FALSE;
		dcb.fDtrControl		= DTR_CONTROL_ENABLE;
		dcb.fDsrSensitivity	= 0;
		dcb.fOutX			= 0;
		dcb.fInX			= 0;
		dcb.fErrorChar		= 0;
		dcb.fNull			= 0;
		dcb.fRtsControl		= RTS_CONTROL_HANDSHAKE;
		dcb.fAbortOnError	= 0;
		dcb.ByteSize		= 8;
		dcb.Parity			= NOPARITY;
		dcb.StopBits		= ONESTOPBIT;

		SetCommState (gCommHandle, &dcb);

		{
			CEnableFullAccess	munge;	// Remove blocks on memory access.

			Platform::SetBaud (READ_REGISTER (uBaud));
			Platform::SetCTS (READ_REGISTER (uTransmit));
		}

		COMMTIMEOUTS	ctmo;

		ctmo.ReadIntervalTimeout			= 50;
		ctmo.ReadTotalTimeoutMultiplier		= 0;
		ctmo.ReadTotalTimeoutConstant		= 100;
		ctmo.WriteTotalTimeoutMultiplier	= 0;
		ctmo.WriteTotalTimeoutConstant		= 0;

		SetCommTimeouts (gCommHandle, &ctmo);

		// Create the threads that will interact with the port.  Make
		// them high priority threads.

		DWORD tid;
		HANDLE commRead = CreateThread (
								NULL,		// security descriptor
								0,			// stack size
								CommRead,	// thread function
								NULL,		// parameter to thread function
								0,			// creation flags
								&tid);		// address of thread ID

		HANDLE commWrite = CreateThread (
								NULL,		// security descriptor
								0,			// stack size
								CommWrite,	// thread function
								NULL,		// parameter to thread function
								0,			// creation flags
								&tid);		// address of thread ID

		SetThreadPriority (commRead, THREAD_PRIORITY_HIGHEST);
		SetThreadPriority (commWrite, THREAD_PRIORITY_HIGHEST);

		CloseHandle (commRead);
		CloseHandle (commWrite);
	}
}


// ---------------------------------------------------------------------------
//		 Platform::CloseSerialPort
// ---------------------------------------------------------------------------

bool Platform::CloseSerialPort (void)
{
	if (gCommHandle)
	{
		// Set gCommHandle to NULL now. That will cause the CommRead
		// thread to quit after we close the handle.

		HANDLE	commHandle = gCommHandle;
		gCommHandle = NULL;

		CloseHandle (commHandle);

		return true;
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 Platform::SetBaud
// ---------------------------------------------------------------------------

void Platform::SetBaud (uae_u16 newRate)
{
	if (gCommHandle)
	{
		DCB		dcb;
		dcb.DCBlength = sizeof (dcb);
		GetCommState (gCommHandle, &dcb);

		// Baud rate is sysClockFreq / preScaler / divider / 16

		dcb.BaudRate = LowMem_GetGlobal(sysClockFreq) /
			(65 - (newRate & hwr328UBaudPrescaler)) /
			(1 << ((newRate & hwr328UBaudDivider) >> hwr328UBaudDivideBitOffset)) /
			16;

		// The above calculation _should_ be right, but it's not. For a rate of
		// 9600 baud, it coughs up 9595.  The values Greg worked out seem to work
		// better.

		dcb.BaudRate = 1036800 /
			(1 << ((newRate & hwr328UBaudDivider) >> hwr328UBaudDivideBitOffset)) /
			(65 - (newRate & hwr328UBaudPrescaler));

		SetCommState (gCommHandle, &dcb);
	}
}


// ---------------------------------------------------------------------------
//		 Platform::SetCTS
// ---------------------------------------------------------------------------

void Platform::SetCTS (uae_u16 uTransmit)
{
	if (gCommHandle)
	{
		DCB		dcb;
		dcb.DCBlength = sizeof (dcb);
		GetCommState (gCommHandle, &dcb);

		bool	useCTS = (uTransmit & hwr328UTransmitIgnoreCTS) == 0;

		dcb.fOutxCtsFlow = useCTS;
		dcb.fRtsControl = useCTS ? RTS_CONTROL_DISABLE : RTS_CONTROL_HANDSHAKE;

		SetCommState (gCommHandle, &dcb);
	}
}


// ---------------------------------------------------------------------------
//		 Platform::WriteSerialByte
// ---------------------------------------------------------------------------

void Platform::WriteSerialByte (uae_u8 outByte)
{
	gCommWriteQueue.Put (outByte);
}


// ---------------------------------------------------------------------------
//		 Platform::ReadSerialByte
// ---------------------------------------------------------------------------

uae_u8 Platform::ReadSerialByte (void)
{
	return gCommReadQueue.Get ();
}


// ---------------------------------------------------------------------------
//		 Platform::IncomingSerialBytes
// ---------------------------------------------------------------------------

uae_u32 Platform::IncomingSerialBytes (void)
{
	return gCommReadQueue.GetUsed ();
}


// ---------------------------------------------------------------------------
//		 Platform::OutgoingSerialBytes
// ---------------------------------------------------------------------------

uae_u32 Platform::OutgoingSerialBytes (void)
{
	return gCommWriteQueue.GetUsed ();
}


// ---------------------------------------------------------------------------
//		 CommRead
// ---------------------------------------------------------------------------
// This function sits in its own thread, waiting for data to show up in the
// serial port.  If data arrives, this function plucks it out and stores it
// in a thread-safe queue.

DWORD __stdcall CommRead (void*)
{
	HANDLE hOverlappedEvent = CreateEvent (NULL, TRUE, FALSE, NULL);

	while (gCommHandle)
	{
		DWORD		error;
		OVERLAPPED	ov;
		memset (&ov, 0, sizeof(ov));
		ov.hEvent = hOverlappedEvent;

		BYTE buf[16];
		DWORD n = gCommReadQueue.GetFree ();

		if (n == 0)
		{
			// No space for new data, so don't read any

			Sleep(10);
			continue;
		}

		if (ReadFile (gCommHandle, buf, n, &n, &ov) ||
			GetLastError () == ERROR_IO_PENDING && GetOverlappedResult (gCommHandle, &ov, &n, TRUE))
		{
			if (n > 0)
			{
				for (DWORD ii = 0; ii < n; ++ii)
				{
					gCommReadQueue.Put (buf[ii]);
				}
			}
		}
	}

	CloseHandle (hOverlappedEvent);

	return 0;
}


// ---------------------------------------------------------------------------
//		 CommWrite
// ---------------------------------------------------------------------------
// This function sits in its own thread, waiting for data to show up in the
// write queue.  If data arrives, this function plucks it out and sends it
// out to the port.

DWORD __stdcall CommWrite (void*)
{
	HANDLE hOverlappedEvent = CreateEvent (NULL, TRUE, FALSE, NULL);

	while (gCommHandle)
	{
		if (!gCommWriteQueue.WaitForDataAvailable (1000))
		{
			continue;
		}

		BYTE buf[8];
		DWORD n = 0;

		while (n < sizeof (buf) && gCommWriteQueue.GetUsed () > 0)
		{
			buf[n++] = gCommWriteQueue.Get ();
		}

		if (n == 0)
		{
			// Never mind, there really wasn't any data in the buffer
			// (see note in TByteQueuePrivate::Put).
			
			continue;
		}

		OVERLAPPED ov;
		memset (&ov, 0, sizeof (ov));
		ov.hEvent = hOverlappedEvent;

		if (!WriteFile (gCommHandle, &buf, n, &n, &ov) && GetLastError () == ERROR_IO_PENDING)
		{
			GetOverlappedResult (gCommHandle, &ov, &n, TRUE);
		}
	}

	CloseHandle (hOverlappedEvent);

	return 0;
}


// ---------------------------------------------------------------------------
//		 TByteQueue::TByteQueue
// ---------------------------------------------------------------------------

TByteQueue::TByteQueue () :
	fImplementation (new TByteQueuePrivate)
{
}


// ---------------------------------------------------------------------------
//		 TByteQueue::~TByteQueue
// ---------------------------------------------------------------------------

TByteQueue::~TByteQueue ()
{
	delete fImplementation;
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Put
// ---------------------------------------------------------------------------

void TByteQueue::Put (uae_u8 b)
{
	fImplementation->Put (b);
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Get
// ---------------------------------------------------------------------------

uae_u8 TByteQueue::Get (void)
{
	return fImplementation->Get ();
}


// ---------------------------------------------------------------------------
//		 TByteQueue::GetUsed
// ---------------------------------------------------------------------------

int TByteQueue::GetUsed (void)
{
	return fImplementation->GetUsed ();
}


// ---------------------------------------------------------------------------
//		 TByteQueue::GetFree
// ---------------------------------------------------------------------------

int TByteQueue::GetFree (void)
{
	return fImplementation->GetFree ();
}


// ---------------------------------------------------------------------------
//		 TByteQueue::WaitForDataAvailable
// ---------------------------------------------------------------------------

bool TByteQueue::WaitForDataAvailable (int timeoutms)
{
	return fImplementation->WaitForDataAvailable (timeoutms);
}


// ---------------------------------------------------------------------------
//		 TByteQueuePrivate::TByteQueuePrivate
// ---------------------------------------------------------------------------

TByteQueuePrivate::TByteQueuePrivate ()
{
	fUsedCount = 0;
	fFreeCount = QUEUE_SIZE;

	fDataAvailable = CreateEvent (NULL, TRUE, FALSE, NULL);

	fHead = 0;
	fTail = 0;
}


// ---------------------------------------------------------------------------
//		 TByteQueuePrivate::~TByteQueuePrivate
// ---------------------------------------------------------------------------

TByteQueuePrivate::~TByteQueuePrivate ()
{
	CloseHandle (fDataAvailable);
	fDataAvailable = NULL;
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Put
// ---------------------------------------------------------------------------

void TByteQueuePrivate::Put (uae_u8 b)
{
	// If the queue is full, just discard the data byte.

	if (fUsedCount >= QUEUE_SIZE)
	{
		return;
	}

	fBuffer[fHead] = b;
	fHead = (fHead + 1) % QUEUE_SIZE;

	InterlockedIncrement(&fUsedCount);
	InterlockedDecrement(&fFreeCount);

	// Tell clients that there may be new data in the buffer.

	// A possible race condition exists: Suppose thread A is at this point
	// in the code, and has just put one byte into an otherwise empty
	// buffer. If thread B then executes TByteQueuePrivate::Get, the byte
	// of data will be removed from the queue. Then thread A will set the
	// fDataAvailable event, when there really is no data available. This
	// is not a problem if clients check the amount of data actually available
	// in the queue when TByteQueue::WaitForDataAvailable returns, rather than
	// assuming there definitely is data.

	SetEvent (fDataAvailable);
}


// ---------------------------------------------------------------------------
//		 TByteQueue::Get
// ---------------------------------------------------------------------------

uae_u8 TByteQueuePrivate::Get ()
{
	// If the queue is empty, return a null byte (this shouldn't happen,
	// because the caller should always call GetUsed before Get).

	if (fUsedCount == 0)
	{
		return 0;
	}

	// If this is the last byte in the queue, reset the
	// fDataAvailable event _before_ removing the byte.

	if (fUsedCount == 1)
	{
		ResetEvent (fDataAvailable);
	}

	uae_u8	r = fBuffer[fTail];
	fTail = (fTail + 1) % QUEUE_SIZE;

	InterlockedIncrement(&fFreeCount);
	InterlockedDecrement(&fUsedCount);

	return r;
}


// ---------------------------------------------------------------------------
//		 TByteQueue::WaitForDataAvailable
// ---------------------------------------------------------------------------

bool TByteQueuePrivate::WaitForDataAvailable (int timeoutms)
{
	return WaitForSingleObject (fDataAvailable, timeoutms) == WAIT_OBJECT_0;
}

